from operator import attrgetter
from typing import List, Optional, Callable
from iobes import SpanEncoding, Span, TokenFunction, SpanFormat, IOB, BIO, IOBES, BILOU, BMEOW
from iobes.utils import safe_get, extract_type


def sort_spans(spans: List[Span]) -> List[Span]:
    """Sort a list of spans ordered by where they start.

    The idea of ordering for spans is that the earlier in the tags the span
    appears (where the span starts) the earlier it will appear in the sorted
    list of spans.

    Args:
        spans: The spans to sort.

    Returns:
        The spans in sorted order. so the earlier in the tag sequence they start the
        earlier they are in this list.
    """
    return sorted(spans, key=attrgetter("start"))


def make_blanks(spans: List[Span], length: Optional[int] = None, fill: str = TokenFunction.OUTSIDE) -> List[str]:
    """Create a list of outside tags that we can populate with tags generated from spans.

    Args:
        spans: The list of spans that will eventually be used to populate the tags.
        length: A pre-specified length for the list of empty tags.
        fill: The value that will be used to populate the list of tags.

    Returns:
        A list of outside tags.
    """
    length = tags_length_from_spans(spans) if length is None else length
    return [fill for _ in range(length)]


def tags_length_from_spans(spans: List[Span]) -> int:
    """Get the length of the list of tags that would be needed to contain all the spans.

    To get a list of tags that are long enough to contain all the spans we need find
    the tag with the largest end index.

    Args:
        spans: The list of spans we need to contain

    Returns:
        The length of the tag list needed.
    """
    return max(spans, key=attrgetter("end")).end


def make_tag(token_function: str, span_type: str, delimiter: str = "-") -> str:
    """Create a tag from a token function and a span type.

    Args:
        token_function: The token function for the tag, it is the first part of the tag.
        span_type: The type of the span this tag is part of it. It is the second part
            of the tag.
        delimiter: A separator character (or sequence of characters) that separate the
            `token_function` and the `span_type`.

    Returns:
        The created tag.
    """
    return f"{token_function}{delimiter}{span_type}"


def write_tags(spans: List[Span], span_type: SpanEncoding, length: Optional[int] = None):
    tags = make_blanks(spans, length)
    if span_type is SpanEncoding.IOB:
        return write_iob_tags(span, length)
    if span_type is SpanEncoding.BIO:
        return _write_tags(spans, BIO, tags)
    if span_type is SpanEncoding.IOBES:
        return _write_tags(spans, IOBES, tags)
    if span_type is SpanEncoding.BILOU:
        return _write_tags(spans, BILOU, tags)
    if span_type is SpanEncoding.BMEOW or span_type is SpanEncoding.BMEWO:
        return _write_tags(spans, BMEOW, tags)
    if span_type is SpanEncoding.TOKEN:
        return _write_tags(spans, TOKEN, tags)
    raise ValueError(f"Unknown SpanEncoding scheme, got: `{span_type}`")


def write_iob_tags(spans: Span, tags: Optional[List[str]] = None, length: Optional[int] = None) -> List[str]:
    """This is a special case because the IOB tags are contextual."""
    spans = sort_spans(spans)
    tags = make_blanks(spans, length) if tags is None else tags
    if not spans:
        return tags
    # The first span can never start with a `B` because it is first and therefore can't follow a
    # span of the same type.
    for token in spans[0].tokens:
        tags[token] = f"{TokenFunction.INSIDE}-{spans[0].type}"
    for prev, span in zip(spans, spans[1:]):
        for token in span.tokens:
            tags[token] = f"{TokenFunction.INSIDE}-{span.type}"
        if prev.end == span.start and prev.type == span.type:
            tags[span.start] = f"{TokenFunction.BEGIN}-{span.type}"
    return tags


def _write_tags(spans: Span, span_format: SpanFormat, tags: Optional[List[str]] = None, length: Optional[int] = None):
    spans = sort_spans(spans)
    tags = make_blanks(spans, length) if tags is None else tags
    for span in spans:
        if span.end == span.start + 1:
            tags[span.start] = make_tag(span_format.SINGLE, span.type)
            continue
        tags[span.start] = make_tag(span_format.BEGIN, span.type)
        tags[span.end - 1] = make_tag(span_format.END, span.type)
        for token in span.tokens[1:-1]:
            tags[token] = make_tag(span_format.INSIDE, span.type)
    return tags


def write_bio_tags(spans: Span, tags: Optional[List[str]] = None, length: Optional[int] = None) -> List[str]:
    return _write_tags(spans, BIO, tags, length)


def write_iobes_tags(spans: Span, tags: Optional[List[str]] = None, length: Optional[int] = None) -> List[str]:
    return _write_tags(spans, IOBES, tags, length)


def write_bilou_tags(spans: Span, tags: Optional[List[str]] = None, length: Optional[int] = None) -> List[str]:
    return _write_tags(spans, BILOU, tags, length)


def write_bmeow_tags(spans: Span, tags: Optional[List[str]] = None, length: Optional[int] = None) -> List[str]:
    return _write_tags(spans, BMEOW, tags, length)


write_bmewo_tags = write_bmeow_tags



from typing import List, Tuple, Callable
from iobes import Span, SpanEncoding, SpanFormat, Error, TokenFunction, LOGGER, IOB, BIO, IOBES, BILOU, BMEOW
from iobes.utils import extract_function, extract_type, safe_get


def parse_spans(seq: List[str], span_type: SpanEncoding) -> List[Span]:
    return parse_spans_with_errors(seq, span_type)[0]


def parse_spans_with_errors(seq: List[str], span_type: SpanEncoding) -> Tuple[List[Span], List[Error]]:
    if span_type is SpanEncoding.IOB:
        return parse_spans_iob_with_errors(seq)
    if span_type is SpanEncoding.BIO:
        return parse_spans_bio_with_errors(seq)
    if span_type is SpanEncoding.IOBES:
        return parse_spans_iobes_with_errors(seq)
    if span_type is SpanEncoding.BILOU:
        return parse_spans_bilou_with_errors(seq)
    if span_type is SpanEncoding.BMEOW or span_type is SpanEncoding.BMEWO:
        return parse_spans_bmeow_with_errors(seq)
    if span_type is SpanEncoding.TOKEN:
        return parse_spans_token_with_errors(seq)
    raise ValueError(f"Unknown SpanEncoding scheme, got: `{span_type}`")


def parse_spans_token(seq: List[str]) -> List[Span]:
    return parse_spans_token_with_errors(seq)[0]


def parse_spans_token_with_errors(seq: List[str]) -> Tuple[List[Span], List[Error]]:
    return [Span(type=t, start=i, end=i + 1, tokens=(i,)) for i, t in enumerate(seq)], []


def parse_spans_iob(seq: List[str]) -> List[Span]:
    return parse_spans_iob_with_errors(seq)[0]


def parse_spans_iob_with_errors(seq: List[str]) -> Tuple[List[Span], List[Error]]:
    errors = []
    spans = []
    # This tracks the type of the span we are currently building
    span = None
    # This tracks the tokens that make up the span we are building
    tokens = []
    for i, s in enumerate(seq):
        func = extract_function(s)
        _type = extract_type(s)
        # A `B` ends a current span but starts a new one
        if func == TokenFunction.BEGIN:
            prev_type = extract_type(seq[i - 1]) if i > 0 else None
            # In `iob` `B` is only allowed to mark the boundary between to spans of the same type that touch
            # `B` isn't allowed to arbitrary start and entity which would happen when `B` is the first token
            # or the last token was an outside
            if i == 0 or prev_type == TokenFunction.OUTSIDE:
                LOGGER.warning("Invalid label: `B` starting an entity at %d", i)
                errors.append(Error(i, "Illegal Start", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
            # If the previous type isn't the same as our type we should have just used an `I` to transition
            elif prev_type != _type:
                LOGGER.warning("Invalid label: `B` starting and entity after a %s at %d", prev_type, i)
                errors.append(Error(i, "Illegal Transition", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
            # If there is a span getting built save it out.
            if span is not None:
                spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
            # Create a new span starting with this B
            span = _type
            tokens = [i]
        # An `I` will continue a span when the types match and force a new one otherwise
        elif func == TokenFunction.INSIDE:
            # There is already a span being build
            if span is not None:
                # If we match types are are a continuation of that span
                if span == _type:
                    tokens.append(i)
                # If we don't match types then we are starting a new span. Save old and start a new one.
                else:
                    spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
                    span = _type
                    tokens = [i]
            # This I starts a new entity
            else:
                span = _type
                tokens = [i]
        # An `O` will end an entity being built
        else:
            # If a span was being made cut it here and save the span out.
            if span is not None:
                spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
            span = None
            tokens = []
    # If we fell off the end save the span that was being made
    if span is not None:
        spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
    return spans, errors


def parse_spans_bio(seq: List[str]) -> List[Span]:
    return parse_spans_bio_with_errors(seq)[0]


def parse_spans_bio_with_errors(seq: List[str]) -> Tuple[List[Span], List[Error]]:
    errors = []
    spans = []
    # This tracks the type of the span we are building out
    span = None
    # This tracks the tokens of the span we are building out
    tokens = []
    for i, s in enumerate(seq):
        func = extract_function(s)
        _type = extract_type(s)
        # A `B` ends a span and starts a new one
        if func == BIO.BEGIN:
            # Save out the old span
            if span is not None:
                spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
            # Start the new span
            span = _type
            tokens = [i]
        # An `I` will continue a span when types match and start a new one otherwise.
        elif func == BIO.INSIDE:
            # A span is already being built
            if span is not None:
                # The types match so we just add to the current span
                if span == _type:
                    tokens.append(i)
                # Types mismatch so create a new span
                else:
                    # Log error from type mismatch
                    LOGGER.warning("Illegal Label: I doesn't match previous token at %d", i)
                    errors.append(Error(i, "Illegal Transition", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
                    # Save out the previous span
                    spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
                    # Start a new span
                    span = _type
                    tokens = [i]
            # No span was being build so start a new one with this I
            else:
                # Log error from starting with I
                LOGGER.warning("Illegal Label: starting a span with `I` at %d", i)
                errors.append(Error(i, "Illegal Start", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
                span = _type
                tokens = [i]
        # An `O` will cut off a span being built out.
        else:
            if span is not None:
                spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
            # Set so no span is being built
            span = None
            tokens = []
    # If we fell off the end so save the entity that we were making.
    if span is not None:
        spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
    return spans, errors


def parse_spans_with_end(seq: List[str], span_format: SpanFormat) -> List[Span]:
    return parse_spans_with_end_with_errors(seq, span_format)[0]


def parse_spans_with_end_with_errors(seq: List[str], span_format: SpanFormat) -> Tuple[List[Span], List[Error]]:
    errors = []
    spans = []
    # The type of the span we are building
    span = None
    # The tokens of the span we are building
    tokens = []
    for i, s in enumerate(seq):
        func = extract_function(s)
        _type = extract_type(s)
        # A `B` ends any current span and starts a new span
        if func == span_format.BEGIN:
            if span is not None:
                # There was a previously active span, This is an error, the span should have been closed by
                # either an `E` or and `S` before starting a new one.
                if i > 0:
                    prev_func = extract_function(seq[i - 1])
                    if prev_func not in (span_format.END, span_format.SINGLE):
                        LOGGER.warning("Illegal Label: `%s` ends span at %d", prev_func, i - 1)
                        errors.append(Error(i - 1, "Illegal End", safe_get(seq, i - 1), safe_get(seq, i - 2), s))
                spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
            span = _type
            tokens = [i]
            # Checking if this `B` causes errors.
            if i < len(seq) - 1:
                next_func = extract_function(seq[i + 1])
                # Look ahead to see if `B` token should actual be can `S` because it is only a single token
                # We only check for `B`, `S` and `O` because an illegal transition to an `I` or `E` will get
                # warned when we actually process that token
                if next_func in (span_format.BEGIN, span_format.SINGLE, TokenFunction.OUTSIDE):
                    LOGGER.warning("Illegal Label: Single `B` token span at %d", i)
                    errors.append(Error(i, "Illegal Single", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
            # A `B` as the last token is an error because it would result in a single span of a `B`
            elif i == len(seq) - 1:
                LOGGER.warning("Illegal Label: `B` as final token %d", i)
                errors.append(Error(i, "Illegal Final", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
        # A `S` ends any active span and creates a new single token span
        elif func == span_format.SINGLE:
            # There was a previously active span, This is an error, the span should have been closed by
            # either an `E` or and `S` before starting a new one.
            if span is not None:
                if i > 0:
                    prev_func = extract_function(seq[i - 1])
                    if prev_func not in (span_format.END, span_format.SINGLE):
                        LOGGER.warning("Illegal Label: `%s` ends span at %d", prev_func, i - 1)
                        errors.append(Error(i - 1, "Illegal End", safe_get(seq, i - 1), safe_get(seq, i - 2), s))
                # Flush this current span
                spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
            # Create a new span that covers this `S`
            spans.append(Span(_type, start=i, end=i + 1, tokens=(i,)))
            # Set the active span to None
            span = None
            tokens = []
        # An `I` will continue a span when the types match and start a new one otherwise.
        elif func == span_format.INSIDE:
            if span is not None:
                # Continue the entity
                if _type == span:
                    tokens.append(i)
                # Out types mismatch, save the current span and start a new one
                else:
                    LOGGER.warning("Illegal Label: `I` doesn't match previous token at %d", i)
                    errors.append(Error(i, "Illegal Transition", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
                    spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
                    span = _type
                    tokens = [i]
            # There was no previous entity we start one with this `I` but this is an error
            else:
                LOGGER.warning("Illegal Label: starting a span with `I` at %d", i)
                errors.append(Error(i, "Illegal Start", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
                span = _type
                tokens = [i]
            # Look ahead to see if this `I` is the last token. This will causes an illegal span because we
            # won't close the span so log this error.
            if i == len(seq) - 1:
                LOGGER.warning("Illegal Label: `I` as final token at %d", i)
                errors.append(Error(i, "Illegal Final", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
        # An `E` will close the currently active span if the type matches. Otherwise we close the current span,
        # create a new span, and immediately close it because we are an `E`
        elif func == span_format.END:
            if span is not None:
                # Type matches to close the span correctly
                if _type == span:
                    tokens.append(i)
                    spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
                    span = None
                    tokens = []
                # Type mismatch
                else:
                    # Log an error that the `E` doesn't match
                    LOGGER.warning("Illegal Label: `E` doesn't match previous token at %d", i)
                    errors.append(Error(i, "Illegal Transition", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
                    # Save out the active span
                    spans.append(Span(span, start=tokens[0], end=i, tokens=tuple(tokens)))
                    # Save out the new span this `E` opens and closes
                    spans.append(Span(_type, start=i, end=i + 1, tokens=(i,)))
                    # Set the active span to None
                    span = None
                    tokens = []
            # There was no span so start and end it with this `E`
            else:
                LOGGER.warning("Illegal Label: starting a span with `E` at %d", i)
                errors.append(Error(i, "Illegal Start", s, safe_get(seq, i - 1), safe_get(seq, i + 1)))
                spans.append(Span(_type, start=i, end=i + 1, tokens=(i,)))
                span = None
                tokens = []
        # An `O` cuts off the active entity
        else:
            # There was a previously active span, This is an error, the span should have been closed by
            # either an `E` or and `S` before having an O
            if span is not None:
                if i > 0:
                    prev_func = extract_function(seq[i - 1])
                    if prev_func not in (span_format.END, span_format.SINGLE):
                        LOGGER.warning("Illegal Label: `%s` ends span at %d", prev_func, i - 1)
                        errors.append(Error(i - 1, "Illegal End", safe_get(seq, i - 1), safe_get(seq, i - 2), s))
                spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
                span = None
                tokens = []
    if span is not None:
        # There was an active entity that fell off the end of the sequence. This should be an error because
        # it means that the span hasn't ended with an `E` or an `S` but we catch these errors by looking
        # ahead in the B or I section instead if doing it here.
        spans.append(Span(span, start=tokens[0], end=tokens[-1] + 1, tokens=tuple(tokens)))
        span = None
        tokens = []
    return spans, errors


def parse_spans_iobes(seq: List[str]) -> List[Span]:
    return parse_spans_iobes_with_errors(seq)[0]


def parse_spans_iobes_with_errors(seq: List[str]) -> Tuple[List[Span], List[Error]]:
    return parse_spans_with_end_with_errors(seq, IOBES)


def parse_spans_bilou(seq: List[str]) -> List[Span]:
    return parse_spans_with_end(seq, BILOU)


def parse_spans_bilou_with_errors(seq: List[str]) -> Tuple[List[Span], List[Error]]:
    return parse_spans_with_end_with_errors(seq, BILOU)


def parse_spans_bmeow(seq: List[str]) -> List[Span]:
    return parse_spans_with_end(seq, BMEOW)


parse_spans_bmewo = parse_spans_bmeow


def parse_spans_bmeow_with_errors(seq: List[str]) -> Tuple[List[Span], List[Error]]:
    return parse_spans_with_end_with_errors(seq, BMEOW)


parse_spans_bmewo_with_errors = parse_spans_bmeow_with_errors


def validate_tags(tags: List[str], span_type: SpanEncoding) -> bool:
    """Check for errors in a tag scheme.

    Args:
        tags: The tags we are parsing.
        span_type: The span encoding scheme we have used.

    Raises:
        ValueError: If the span encoding scheme isn't recognized.

    Returns:
        True if the tags don't have any formatting errors, False otherwise.
    """
    if span_type is SpanEncoding.IOB:
        return validate_tags_iob(tags)
    if span_type is SpanEncoding.BIO:
        return validate_tags_bio(tags)
    if span_type is SpanEncoding.IOBES:
        return validate_tags_iobes(tags)
    if span_type is SpanEncoding.BILOU:
        return validate_tags_bilou(tags)
    if span_type is SpanEncoding.BMEOW or span_type is SpanEncoding.BMEWO:
        return validate_tags_bmeow(tags)
    if span_type is SpanEncoding.TOKEN:
        return validate_tags_token(tags)
    raise ValueError(f"Unknown SpanEncoding Scheme, got: `{span_type}`")


def _validate_tags(parse: Callable[[List[str]], Tuple[List[Span], List[Error]]], tags: List[str]) -> bool:
    """Check for errors in a tag scheme.

    Args:
        parse: A function that parses spans and return spans and errors.
        tags: The tags we are parsing.

    Returns:
        True if the tags don't have any formatting errors, False otherwise.
    """
    _, errors = parse(tags)
    return not errors


def validate_tags_iob(tags: List[str]) -> bool:
    """Check for errors in IOB tags.

    Args:
        tags: The IOB tags we are parsing.

    Returns:
        True if the IOB tags are well-formed, False otherwise.
    """
    return _validate_tags(parse_spans_iob_with_errors, tags)


def validate_tags_bio(tags: List[str]) -> bool:
    """Check for errors in BIO tags.

    Args:
        tags: The BIO tags we are parsing.

    Returns:
        True if the BIO tags are well-formed, False otherwise.
    """
    return _validate_tags(parse_spans_bio_with_errors, tags)


def validate_tags_iobes(tags: List[str]) -> bool:
    """Check for errors in IOBES tags.

    Args:
        tags: The IOBES tags we are parsing.

    Returns:
        True if the IOBES tags are well-formed, False otherwise.
    """
    return _validate_tags(parse_spans_iobes_with_errors, tags)


def validate_tags_bilou(tags: List[str]) -> bool:
    """Check for errors in BILOU tags.

    Args:
        tags: The BILOU tags we are parsing.

    Returns:
        True if the BILOU tags are well-formed, False otherwise.
    """
    return _validate_tags(parse_spans_bilou_with_errors, tags)


def validate_tags_bmeow(tags: List[str]) -> bool:
    """Check for errors in BMEOW tags.

    Args:
        tags: The BMEOW tags we are parsing.

    Returns:
        True if the BMEOW tags are well-formed, False otherwise.
    """
    return _validate_tags(parse_spans_bmeow_with_errors, tags)


def validate_tags_token(tags: List[str]) -> bool:
    """Check for errors in TOKEN tags.

    Note:
        Token tags are not processed into spans so all sequences are valid.

    Args:
        tags: The TOKEN tags we are parsing.

    Returns:
        True
    """
    return True


def validate_tags_bmewo(tags: List[str]) -> bool:
    """Check for errors in BMEWO tags.

    Note:
        Alias for :py:func:`~iobes.parse.validate_labels_bmeow`

    Args:
        tags: The BMEWO tags we are parsing.

    Returns:
        True if the BMEWO tags are well-formed, False otherwise.
    """
    return validate_tags_bmeow(tags)



from itertools import chain
from typing import List, Tuple, Callable
from iobes import TokenFunction, Span, Error
from iobes.parse import (
    parse_spans_iob_with_errors,
    parse_spans_bio_with_errors,
    parse_spans_iobes_with_errors,
    parse_spans_bilou_with_errors,
    parse_spans_bmeow_with_errors,
)
from iobes.write import (
    write_iob_tags,
    write_bio_tags,
    write_iobes_tags,
    write_bilou_tags,
    write_bmeow_tags,
)


def convert_tags(
    tags: List[str],
    parse_function: Callable[[List[str]], Tuple[List[Span], List[Error]]],
    write_function: Callable[[List[Span], int], List[str]],
) -> List[str]:
    """Convert tags from one format to another.

    Args:
        tags: The tags that we are converting.
        parse_function: A function that parses tags into spans.
        write_function: A function the turns spans into a list of tags.

    Raises:
        ValueError: If there were errors in the tag formatting.

    Returns:
        The list of tags in the new format.
    """
    spans, errors = parse_function(tags)
    if errors:
        error_string = "\n".join(str(e) for e in errors)
        raise ValueError(f"Found errors in the tag sequence, cannot be converted. Errors: {error_string}")
    return write_function(spans, length=len(tags))


def iob_to_bio(tags: List[str]) -> List[str]:
    """Convert IOB tags to the BIO format.

    Args:
        tags: The IOB tags we are converting

    Raises:
        ValueError: If there were errors in the IOB formatting of the input.

    Returns:
        Tags that produce the same spans in the BIO format.
    """
    return convert_tags(tags, parse_spans_iob_with_errors, write_bio_tags)


def iob_to_iobes(tags: List[str]) -> List[str]:
    """Convert IOB tags to the IOBES format.

    Args:
        tags: The IOB tags we are converting

    Raises:
        ValueError: If there were errors in the IOB formatting of the input.

    Returns:
        Tags that produce the same spans in the IOBES format.
    """
    return bio_to_iobes(iob_to_bio(tags))


def iob_to_bilou(tags: List[str]) -> List[str]:
    """Convert IOB tags to the BILOU format.

    Args:
        tags: The IOB tags we are converting

    Raises:
        ValueError: If there were errors in the IOB formatting of the input.

    Returns:
        Tags that produce the same spans in the BILOU format.
    """
    return iobes_to_bilou(iob_to_iobes(tags))


def iob_to_bmeow(tags: List[str]) -> List[str]:
    """Convert IOB tags to the BMEOW format.

    Args:
        tags: The IOB tags we are converting

    Raises:
        ValueError: If there were errors in the IOB formatting of the input.

    Returns:
        Tags that produce the same spans in the BMEOW format.
    """
    return iobes_to_bmeow(iob_to_iobes(tags))


def iob_to_bmewo(tags: List[str]) -> List[str]:
    """Convert IOB tags to the BMEWO format.

    Note:
        Alias for :py:func:`~iobes.convert.iob_to_bmeow`.

    Args:
        tags: The IOB tags we are converting

    Raises:
        ValueError: If there were errors in the IOB formatting of the input.

    Returns:
        Tags that produce the same spans in the BMEOW format.
    """
    return iob_to_bmeow(tags)


def bio_to_iob(tags: List[str]) -> List[str]:
    """Convert BIO tags to the IOB format.

    Args:
        tags: The BIO tags we are converting

    Raises:
        ValueError: If there were errors in the BIO formatting of the input.

    Returns:
        Tags that produce the same spans in the IOB format.
    """
    return convert_tags(tags, parse_spans_bio_with_errors, write_iob_tags)


def bio_to_iobes(tags: List[str]) -> List[str]:
    """Convert BIO tags to the IOBES format.

    Args:
        tags: The BIO tags we are converting

    Raises:
        ValueError: If there were errors in the BIO formatting of the input.

    Returns:
        Tags that produce the same spans in the IOBES format.
    """
    return convert_tags(tags, parse_spans_bio_with_errors, write_iobes_tags)


def bio_to_bilou(tags: List[str]) -> List[str]:
    """Convert BIO tags to the BILOU format.

    Args:
        tags: The BIO tags we are converting

    Raises:
        ValueError: If there were errors in the BIO formatting of the input.

    Returns:
        Tags that produce the same spans in the BILOU format.
    """
    return iobes_to_bilou(bio_to_iobes(tags))


def bio_to_bmeow(tags: List[str]) -> List[str]:
    """Convert BIO tags to the BMEOW format.

    Args:
        tags: The BIO tags we are converting

    Raises:
        ValueError: If there were errors in the BIO formatting of the input.

    Returns:
        Tags that produce the same spans in the BMEOW format.
    """
    return iobes_to_bmeow(bio_to_iobes(tags))


def bio_to_bmewo(tags: List[str]) -> List[str]:
    """Convert BIO tags to the BMEWO format.

    Note:
        Alias for :py:func:`~iobes.convert.bio_to_bmeow`

    Args:
        tags: The BIO tags we are converting

    Raises:
        ValueError: If there were errors in the BIO formatting of the input.

    Returns:
        Tags that produce the same spans in the BMEWO format.
    """
    return bio_to_bmeow(tags)


def iobes_to_iob(tags: List[str]) -> List[str]:
    """Convert IOBES tags to the IOB format.

    Args:
        tags: The IOBES tags we are converting

    Raises:
        ValueError: If there were errors in the IOBES formatting of the input.

    Returns:
        Tags that produce the same spans in the IOB format.
    """
    return bio_to_iob(iobes_to_bio(tags))


def iobes_to_bio(tags: List[str]) -> List[str]:
    """Convert IOBES tags to the BIO format.

    Args:
        tags: The IOBES tags we are converting

    Raises:
        ValueError: If there were errors in the IOBES formatting of the input.

    Returns:
        Tags that produce the same spans in the BIO format.
    """
    return convert_tags(tags, parse_spans_iobes_with_errors, write_bio_tags)


def iobes_to_bilou(tags: List[str]) -> List[str]:
    """Convert IOBES tags to the BILOU format.

    Args:
        tags: The IOBES tags we are converting

    Raises:
        ValueError: If there were errors in the IOBES formatting of the input.

    Returns:
        Tags that produce the same spans in the BILOU format.
    """
    return convert_tags(tags, parse_spans_iobes_with_errors, write_bilou_tags)


def iobes_to_bmeow(tags: List[str]) -> List[str]:
    """Convert IOBES tags to the BMEOW format.

    Args:
        tags: The IOBES tags we are converting

    Raises:
        ValueError: If there were errors in the IOBES formatting of the input.

    Returns:
        Tags that produce the same spans in the BMEOW format.
    """
    return convert_tags(tags, parse_spans_iobes_with_errors, write_bmeow_tags)


def iobes_to_bmewo(tags: List[str]) -> List[str]:
    """Convert IOBES tags to the BMEWO format.

    Note:
        Alias for :py:func:`~iobes.convert.iobes_to_bmeow`

    Args:
        tags: The IOBES tags we are converting

    Raises:
        ValueError: If there were errors in the IOBES formatting of the input.

    Returns:
        Tags that produce the same spans in the BMEWO format.
    """
    return iobes_to_bmeow(tags)


def bilou_to_iob(tags: List[str]) -> List[str]:
    """Convert BILOU tags to the IOB format.

    Args:
        tags: The BILOU tags we are converting

    Raises:
        ValueError: If there were errors in the BILOU formatting of the input.

    Returns:
        Tags that produce the same spans in the IOB format.
    """
    return convert_tags(tags, parse_spans_bilou_with_errors, write_iob_tags)


def bilou_to_bio(tags: List[str]) -> List[str]:
    """Convert BILOU tags to the BIO format.

    Args:
        tags: The BILOU tags we are converting

    Raises:
        ValueError: If there were errors in the BILOU formatting of the input.

    Returns:
        Tags that produce the same spans in the BIO format.
    """
    return convert_tags(tags, parse_spans_bilou_with_errors, write_bio_tags)


def bilou_to_iobes(tags: List[str]) -> List[str]:
    """Convert BILOU tags to the IOBES format.

    Args:
        tags: The BILOU tags we are converting

    Raises:
        ValueError: If there were errors in the BILOU formatting of the input.

    Returns:
        Tags that produce the same spans in the IOBES format.
    """
    return convert_tags(tags, parse_spans_bilou_with_errors, write_iobes_tags)


def bilou_to_bmeow(tags: List[str]) -> List[str]:
    """Convert BILOU tags to the BMEOW format.

    Args:
        tags: The BILOU tags we are converting

    Raises:
        ValueError: If there were errors in the BILOU formatting of the input.

    Returns:
        Tags that produce the same spans in the BMEOW format.
    """
    return convert_tags(tags, parse_spans_bilou_with_errors, write_bmeow_tags)


def bilou_to_bmewo(tags: List[str]) -> List[str]:
    """Convert BILOU tags to the BMEWO format.

    Note:
        Alias for :py:func:`~iobes.convert.bilou_to_bmeow`

    Args:
        tags: The BILOU tags we are converting

    Raises:
        ValueError: If there were errors in the BILOU formatting of the input.

    Returns:
        Tags that produce the same spans in the BMEWO format.
    """
    return bilou_to_bmeow(tags)


def bmeow_to_iob(tags: List[str]) -> List[str]:
    """Convert BMEOW tags to the IOB format.

    Args:
        tags: The BMEOW tags we are converting

    Raises:
        ValueError: If there were errors in the BMEOW formatting of the input.

    Returns:
        Tags that produce the same spans in the IOB format.
    """
    return convert_tags(tags, parse_spans_bmeow_with_errors, write_iob_tags)


def bmeow_to_bio(tags: List[str]) -> List[str]:
    """Convert BMEOW tags to the BIO format.

    Args:
        tags: The BMEOW tags we are converting

    Raises:
        ValueError: If there were errors in the BMEOW formatting of the input.

    Returns:
        Tags that produce the same spans in the BIO format.
    """
    return convert_tags(tags, parse_spans_bmeow_with_errors, write_bio_tags)


def bmeow_to_iobes(tags: List[str]) -> List[str]:
    """Convert BMEOW tags to the IOBES format.

    Args:
        tags: The BMEOW tags we are converting

    Raises:
        ValueError: If there were errors in the BMEOW formatting of the input.

    Returns:
        Tags that produce the same spans in the IOBES format.
    """
    return convert_tags(tags, parse_spans_bmeow_with_errors, write_iobes_tags)


def bmeow_to_bilou(tags: List[str]) -> List[str]:
    """Convert BMEOW tags to the BILOU format.

    Args:
        tags: The BMEOW tags we are converting

    Raises:
        ValueError: If there were errors in the BMEOW formatting of the input.

    Returns:
        Tags that produce the same spans in the BILOU format.
    """
    return convert_tags(tags, parse_spans_bmeow_with_errors, write_bilou_tags)


def bmewo_to_iob(tags: List[str]) -> List[str]:
    """Convert BMEWO tags to the IOB format.

    Note:
        Alias for :py:func:`~iobes.convert.bmeow_to_iob`

    Args:
        tags: The BMEWO tags we are converting

    Raises:
        ValueError: If there were errors in the BMEWO formatting of the input.

    Returns:
        Tags that produce the same spans in the IOB format.
    """
    return bmeow_to_iob(tags)


def bmewo_to_bio(tags: List[str]) -> List[str]:
    """Convert BMEWO tags to the BIO format.

    Note:
        Alias for :py:func:`~iobes.convert.bmeow_to_bio`

    Args:
        tags: The BMEWO tags we are converting

    Raises:
        ValueError: If there were errors in the BMEWO formatting of the input.

    Returns:
        Tags that produce the same spans in the BIO format.
    """
    return bmeow_to_bio(tags)


def bmewo_to_iobes(tags: List[str]) -> List[str]:
    """Convert BMEWO tags to the IOBES format.

    Note:
        Alias for :py:func:`~iobes.convert.bmeow_to_iobes`

    Args:
        tags: The BMEWO tags we are converting

    Raises:
        ValueError: If there were errors in the BMEWO formatting of the input.

    Returns:
        Tags that produce the same spans in the IOBES format.
    """
    return bmeow_to_iobes(tags)


def bmewo_to_bilou(tags: List[str]) -> List[str]:
    """Convert BMEWO tags to the BILOU format.

    Note:
        Alias for :py:func:`~iobes.convert.bmeow_to_bilou`

    Args:
        tags: The BMEWO tags we are converting

    Raises:
        ValueError: If there were errors in the BMEWO formatting of the input.

    Returns:
        Tags that produce the same spans in the BILOU format.
    """
    return bmeow_to_bilou(tags)



from itertools import chain, permutations
from typing import List, Tuple, NamedTuple, Dict, Set
from iobes import SpanEncoding, TokenFunction, SpanFormat, IOB, BIO, IOBES, BILOU, BMEOW
from iobes.utils import extract_function, extract_type


class Transition(NamedTuple):
    """A transition from one state to another.

    This includes information about whether the transition is legal or not. The legality of
    a transition is dictated by the span encoding scheme used.

    Args:
        source: The state you are starting at.
        target: The state you are going to.
        valid: Is this transition allowed by the encoding scheme?
    """

    source: str
    target: str
    valid: bool


def transitions_legality(
    tags: Set[str], span_type: SpanEncoding, start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    """Get the transition legality for some SpanEncoding format.

    Return a list of transitions and their legality based on the SpanEncoding schemes and
    the types of spans present.

    This is a convenience function that dispatches to span encoding specific implementations
    based on the span_type.

    Note:
        We include special tags that represent the start and end of sequences. These are
        special values that used downstream implementations of things like Conditional
        Random Fields (CRFs) `Lafferty et. al., 2001`__ and helps define constraints about
        what tags are allowed on the first and last token in a sequence. General rules around
        the start symbol is that nothing can transition to the start token and the legal targets
        of a transition from a start symbol is limited by the span encoding scheme. Similarly
        the end token cannot transition into anything else and what can transition to it is
        specified by the encoding scheme.

    Args:
        tags: The tags that we can assign to tokens.
        span_type: The span encoding format we are trying to use. Different formats impose different
            rules about which transitions are legal or not.
        start: A special tag representing the start of all sequences.
        end: A special tag representing the end of all sequences.

    Raises:
        ValueError: If the span encoding scheme isn't recognized.

    Returns:
        The list of transitions.

    __ https://repository.upenn.edu/cgi/viewcontent.cgi?article=1162&context=cis_papers
    """
    if span_type is SpanEncoding.IOB:
        return iob_transitions_legality(tags, start, end)
    if span_type is SpanEncoding.BIO:
        return bio_transitions_legality(tags, start, end)
    if span_type is SpanEncoding.IOBES:
        return iobes_transitions_legality(tags, start, end)
    if span_type is SpanEncoding.BILOU:
        return bilou_transitions_legality(tags, start, end)
    if span_type is SpanEncoding.BMEOW or span_type is SpanEncoding.BMEWO:
        return bmeow_transitions_legality(tags, start, end)
    if span_type is SpanEncoding.TOKEN:
        return token_transitions_legality(tags, start, end)
    raise ValueError(f"Unknown SpanEncoding Scheme, got: `{span_type}`")


def token_transitions_legality(
    tags: Set[str], start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    """Get transition legality when processing tokens.

    Note:
        Token level annotations like Part of Speech Tagging don't have transition constrains
        defined by the span encoding scheme (because there is no span encoding scheme). This
        means that most every transition is allowed. The only illegal transitions are moving
        back to the special start token or leaving the end token.

    Args:
        tags: The tags that we can assign to tokens.
        start: A special tag representing the start of all sequences.
        end: A special tag representing the end of all sequences.

    Returns:
        The list of transitions.
    """
    transitions = []
    for src, tgt in permutations(tags, 2):
        transitions.append(Transition(src, tgt, True))
    for tag in tags:
        transitions.append(Transition(tag, tag, True))
        transitions.append(Transition(start, tag, True))
        transitions.append(Transition(tag, start, False))
        transitions.append(Transition(tag, end, True))
        transitions.append(Transition(end, tag, False))
    return transitions


def iob_transitions_legality(
    tags: Set[str], start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    """Get transition legality when processing `IOB` tags.

    There are a few rules the govern `IOB` tagging. Spans are allowed to begin with an `I-`
    so a lot of the rules other span encoding formats about not transitioning from `O` to
    and `I-` don't apply. The main rules are around the use of the `B-` token. In `IOB` we
    are only allowed to start a token with a `B-` when it is the start of a new span that
    directly follows (touches) a previous span of the same time. This translates into rules
    that `B-` tokens can only follow tags that have the same type (either `B-` or `I-`)

    Args:
        tags: The tags that we can assign to tokens.
        start: A special tag representing the start of all sequences.
        end: A special tag representing the end of all sequences.

    Returns:
        The list of transitions.
    """
    transitions = []
    for src in chain(tags, [start, end]):
        src_func = extract_function(src)
        src_type = extract_type(src)
        for tgt in chain(tags, [start, end]):
            tgt_func = extract_function(tgt)
            tgt_type = extract_type(tgt)
            # Can't transition to start
            if tgt == start:
                transitions.append(Transition(src, tgt, False))
                continue
            # Can't transition from start
            if src == end:
                transitions.append(Transition(src, tgt, False))
                continue
            # Can't transition from start to B because B needs to be between two spans of the same type
            elif src == start:
                if tgt_func == TokenFunction.BEGIN:
                    transitions.append(Transition(src, tgt, False))
                    continue
            elif src_func == TokenFunction.BEGIN:
                # Can only go from B to B of the same type
                if tgt_func == TokenFunction.BEGIN:
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == TokenFunction.INSIDE:
                # Can only go from I to B of the same type
                if tgt_func == TokenFunction.BEGIN:
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == TokenFunction.OUTSIDE:
                # Can't start a span with B unless preceded by another span
                if tgt_func == TokenFunction.BEGIN:
                    transitions.append(Transition(src, tgt, False))
                    continue
            transitions.append(Transition(src, tgt, True))
    return transitions


def bio_transitions_legality(
    tags: Set[str], start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    """Get transition legality when processing `BIO` tags.

    **TODO**

    Args:
        tags: The tags that we can assign to tokens.
        start: A special tag representing the start of all sequences.
        end: A special tag representing the end of all sequences.

    Returns:
        The list of transitions.
    """
    transitions = []
    for src in chain(tags, [start, end]):
        src_func = extract_function(src)
        src_type = extract_type(src)
        for tgt in chain(tags, [start, end]):
            tgt_func = extract_function(tgt)
            tgt_type = extract_type(tgt)
            # Can't transition to start
            if tgt == start:
                transitions.append(Transition(src, tgt, False))
                continue
            # Can't transition from end
            if src == end:
                transitions.append(Transition(src, tgt, False))
                continue
            elif src == start:
                # Can't go from start to an I
                if tgt_func == BIO.INSIDE:
                    transitions.append(Transition(src, tgt, False))
                    continue
            elif src_func == BIO.BEGIN:
                # Can only go from B to I of same type
                if tgt_func == BIO.INSIDE:
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == BIO.INSIDE:
                # Can only go from I to I of same type
                if tgt_func == BIO.INSIDE:
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == TokenFunction.OUTSIDE:
                # Can't start an entity with I
                if tgt_func == BIO.INSIDE:
                    transitions.append(Transition(src, tgt, False))
                    continue
            transitions.append(Transition(src, tgt, True))
    return transitions


def with_end_transitions_legality(
    tags: Set[str], span_format: SpanFormat, start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    """Get transition legality when processing tags when the encoding scheme has a `end` token function.

    Span encoding schemes that have special token prefixes for tokens that are the start, middle, and end
    of a span (and a specific prefix for a token that represents a single token span) have quite a few more
    rule. These can mostly be summed up as spans need to start with the starting prefix and end with the
    ending prefix. What this means that things like the inside tokens can't follow an outside and can't be
    followed by an outside. It also has rules like the beginning token can't be followed by an ending
    token that is a different type.

    Note:
        Several span formats like `IOBES`, `BILOU`, and `BMEOW` are the same except for the value
        of some of the `TokenFunction` (`IOBES` has `E` for the end while `BILOU` has `L`). Other
        than these differences these all behave the same way. This function parses all of these
        formats by comparing to the things like the `SpanFormat.BEGIN` instead of the literal
        string. This is the underlying implementation but the user facing function to get the
        transitions for a specific encoding scheme should be used.

    Args:
        tags: The tags that we can assign to tokens.
        span_format: The `SpanFormat` we are using for these tags.
        start: A special tag representing the start of all sequences.
        end: A special tag representing the end of all sequences.

    Returns:
        The list of transitions.
    """
    transitions = []
    for src in chain(tags, [start, end]):
        src_func = extract_function(src)
        src_type = extract_type(src)
        for tgt in chain(tags, [start, end]):
            tgt_func = extract_function(tgt)
            tgt_type = extract_type(tgt)
            # Can't transition to start
            if tgt == start:
                transitions.append(Transition(src, tgt, False))
                continue
            # Can't transition from end
            if src == end:
                transitions.append(Transition(src, tgt, False))
                continue
            elif src == start:
                # Can't start span with I or E
                if tgt_func in (span_format.INSIDE, span_format.END):
                    transitions.append(Transition(src, tgt, False))
                    continue
            elif src_func == span_format.BEGIN:
                # Can't go from B to B, S, or O because we didn't close the entity
                if tgt_func in (span_format.BEGIN, span_format.SINGLE, TokenFunction.OUTSIDE) or tgt_func == end:
                    transitions.append(Transition(src, tgt, False))
                    continue
                # Can only go from B to I or E of the same type
                elif tgt_func in (span_format.INSIDE, span_format.END):
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func == span_format.INSIDE:
                # Can't from from I to B, S, or O because we didin't close the entity
                if tgt_func in (span_format.BEGIN, span_format.SINGLE, TokenFunction.OUTSIDE) or tgt == end:
                    transitions.append(Transition(src, tgt, False))
                    continue
                # Can only go from I to I or E of the same Type
                elif tgt_func in (span_format.INSIDE, span_format.END):
                    if src_type != tgt_type:
                        transitions.append(Transition(src, tgt, False))
                        continue
            elif src_func in (span_format.END, span_format.SINGLE, TokenFunction.OUTSIDE):
                # Going from outside an entity (or ending it) to one that was inside the entity (I/E) is illegal
                if tgt_func in (span_format.INSIDE, span_format.END):
                    transitions.append(Transition(src, tgt, False))
                    continue
            # Other transitions are allowed
            transitions.append(Transition(src, tgt, True))
    return transitions


def iobes_transitions_legality(
    tags: Set[str], start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    """Get transition legality when processing IOBES tags.

    **TODO**

    Args:
        tags: The tags that we can assign to tokens.
        start: A special tag representing the start of all sequences.
        end: A special tag representing the end of all sequences.

    Returns:
        The list of transitions.
    """
    return with_end_transtitions_legality(tags, IOBES, start, end)


def bilou_transitions_legality(
    tags: Set[str], start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    """Get transition legality when processing `BILOU` tags.

    **TODO**

    Args:
        tags: The tags that we can assign to tokens.
        start: A special tag representing the start of all sequences.
        end: A special tag representing the end of all sequences.

    Returns:
        The list of transitions.
    """
    return with_end_transtitions_legality(tags, BILOU, start, end)


def bmeow_transitions_legality(
    tags: Set[str], start: str = TokenFunction.GO, end: str = TokenFunction.EOS
) -> List[Transition]:
    """Get transition legality when processing BMEOW tags.

    **TODO**

    Args:
        tags: The tags that we can assign to tokens.
        start: A special tag representing the start of all sequences.
        end: A special tag representing the end of all sequences.

    Returns:
        The list of transitions.
    """
    return with_end_transtitions_legality(tags, BMEOW, start, end)


bmewo_transitions_legality = bmeow_transitions_legality


def transitions_to_tuple_map(transitions: List[Transition]) -> Dict[Tuple[str, str], bool]:
    """Convert the list of transitions to a dictionary keyed by the `(source, target)` tuple.

    This data structure is useful when given a pair of states you want to check if the transition
    is legal in O(1) time.

    Args:
        transitions: The list of transitions.

    Returns:
        A dictionary mapping `(source, target)` pairs to the legality of that transition.
    """
    return {(src, tgt): valid for src, tgt, valid in transitions}


def transitions_to_map(transitions: List[Transition]) -> Dict[str, Dict[str, bool]]:
    """Convert the list of transitions into nested dictionaries keyed by the states.

    The data format is a dictionary mapping `source` to a dictionary of `target`. This inner
    dictionary has the legality of the transition as values. For example `result[src][tgt]`
    being `True` means that the transition from src to tgt is valid while a `False` means
    it is illegal.

    This data structure is useful when given a state you want to see the legality of transitions
    from it to other states.

    Args:
        transitions: The list of transitions.

    Returns:
        Nested dictionaries representing the legality of transitions.
    """
    mapping = {}
    for src, tgt, valid in transitions:
        to = mapping.setdefault(src, {})
        to[tgt] = valid
    return mapping


def transitions_to_mask(transitions: List[Transition], vocabulary: Dict[str, int]) -> "np.ndarray":
    """Convert the list of transitions into a mask.

    The starting state is represented by the row index in the mask while the ending state is represented
    by the column index. A value of one in the mask means the transition was legal while a zero means
    it was illegal. For example, `mask[src, tgt] == 1` is means the transition from src to tgt was allowed
    while a zero means it is not.

    This data structure is useful when you have a transition matrix that represents something like the
    probability of transitions between states and you want to zero out the value for illegal transitions.

    Note:
        This function has a dependency on `numpy`. This is an optional dependency for the `iobes`
        library and can installed with `pip install iobes[mask]`.

    Args:
        transitions: The list of transitions.
        vocabulary: A mapping of state name to index. This is used to figure out where to place
            the state value in the mask.

    Returns:
        A mask representing the legal and illegal transitions.
    """
    try:
        import numpy as np
    except ImportError as e:
        LOGGER.critical(
            "Could not import the `numpy` library which is needed to create a transition mask. Use `pip install iobes[mask] to include the optional `numpy` dependency."
        )
        raise e
    mask = np.zeros((len(vocabulary), len(vocabulary)))
    for src, tgt, value in transitions:
        if value:
            mask[vocabulary[src], vocabulary[tgt]] = 1
    return mask



__version__ = "1.4.3"

import logging
from enum import Enum
from typing import NamedTuple, Tuple


LOGGER = logging.getLogger("iobes")


class SpanFormat:
    """A description of a tag format.

    The anatomy of a tag is ``{token-function}-{span-type}``. It has two parts, the second part
    is the type of the span. This is specific to the downstream task and can be things like
    ``PER`` or ``LOC`` for NER or things like ``to_city`` and ``from_city`` for slot filling used in a
    dialogue manager for air travel booking. The first part is the token function. It is generic
    across tasks and it is used when converting per-token labels into spans. Common examples is
    when a tag starts with a ``B-`` we know that it is the beginning of a new span or when a tag
    starts with ``I-`` we know that is is inside of a span.

    Note:
        Some tag formats have special ``end`` and ``single`` values for tags that end a span and
        tags that constitute a whole span in themselves while others don't. The span encoding
        formats that don't have end and single tokens just repeats of the ``inside`` and ``begin``
        attributes respectively.
    """

    BEGIN = None  #: This is the token function that all tags that trigger a new span should have
    INSIDE = None  #: This is the token function that all tags inside of a span should have
    END = None  #: This is the token function that all tags at the end of a span should have
    SINGLE = None  #: This is the token function that all tags that constitute a span of length 1 should have

    def __init__(self):
        """These formats are designed to be singletons with class attributes so stop people from creating an object."""
        raise NotImplementedError("You cannot instantiate a SpanFormat object")


class TokenFunction:
    """Prefixes for tags that are used in decoding.

    In general tags can be broken into two parts, The first is the token function which
    tells you something about how the decoding parser should act when it hits this tag
    and the second half is the type (PER, LOC, etc) of the span.
    """

    OUTSIDE = "O"  #: This tag is not in any span, this is a rare one that is a whole tag, not just a prefix
    BEGIN = "B"  #: This tag starts a span
    INSIDE = "I"  #: This tag is in the middle of a span
    MIDDLE = "M"  #: This tag is in the middle of a span
    END = "E"  #: This tag ends a span
    LAST = "L"  #: This tag ends a span
    SINGLE = "S"  #: This tag by itself represents a span
    UNIT = "U"  #: This tag by itself represents a span
    WHOLE = "W"  #: This tag by itself represents a span
    GO = "<GO>"  #: This tag is a special tag for the beginning of a sequence
    EOS = "<EOS>"  #: This tag is a special tag for the end of a sequence


class IOB(SpanFormat):
    """The original IOB tagging format.

    The first span encoding format proposed in `Ramshaw and Marcus, 1995`_

    This is the only format this is contextual, When two spans for the same type are touching then
    the first token of the second span would be a ``B-`` where as in cases when the first token is
    not following (touching) another span of the same type it would be an ``I-``. So the value of the
    BEGIN tag isn't known without context. The same applies to the SINGLE tag. When a span is a single
    token the prefix will be ``I-`` if it is preceded by no span, or a span of a different type. It
    would use the prefix ``B-`` if the previous span was that same type.

    .. _Ramshaw and Marcus, 1995: https://www.aclweb.org/anthology/W95-0107/
    """

    BEGIN = None  #: The prefix for the beginning of the span in unknown a priori
    INSIDE = TokenFunction.INSIDE  #: The inside of a span is always known.
    END = TokenFunction.INSIDE  #: The end token is always known, it is the same as the inside token.
    SINGLE = None  #: Like the beginning token, the single token span is unknown without the previous span type.


class BIO(SpanFormat):
    """The improved BIO tagging format.

    This is an improvement to the IOB format. All entities, regardless of the value of the previous span,
    start with a ``B-`` token. This is a context independent format because we always know that the first
    token is a ``B-``. There is not special end tag however. Things like an ``O`` and a token of a different
    type trigger the end of the entity.
    """

    BEGIN = TokenFunction.BEGIN
    INSIDE = TokenFunction.INSIDE
    END = TokenFunction.INSIDE
    SINGLE = TokenFunction.BEGIN


class IOBES(SpanFormat):
    """The best tagging format.

    ** TODO ** flesh out

    This format adds an END tag that needs to show up at the end of entities. This format has been shown
    to be better than IOB or BIO (`Ratinov and Roth, 2009`_) and should be used instead.

    .. _Ratinov and Roth, 2009: https://www.aclweb.org/anthology/W09-1119/
    """

    BEGIN = TokenFunction.BEGIN
    INSIDE = TokenFunction.INSIDE
    END = TokenFunction.END
    SINGLE = TokenFunction.SINGLE


class BILOU(SpanFormat):
    """The BILOU format.

    ** TODO ** flesh out

    This is the same as the IOBES format but we just have different values for the END and SINGLE tokens.
    """

    BEGIN = TokenFunction.BEGIN
    INSIDE = TokenFunction.INSIDE
    END = TokenFunction.LAST
    SINGLE = TokenFunction.UNIT


class BMEOW(SpanFormat):
    """The BMEOW format.

    ** TODO ** flesh out

    From `Borthwick, 1999`_

    This is the same as the IOBES format but we just have different values for the INSIDE and SINGLE tokens.

    .. _Borthwick, 1999: https://www.math.nyu.edu/media/mathfin/publications/borthwick_andrew.pdf
    """

    BEGIN = TokenFunction.BEGIN
    INSIDE = TokenFunction.MIDDLE
    END = TokenFunction.END
    SINGLE = TokenFunction.WHOLE


BMEWO = (  #: This is the same as BMEOW and what a lot of people actually call it but having `meow` in it seems better lol.
    BMEOW
)


class TOKEN(SpanFormat):
    """A format to use when processing tokens.

    In this case the tags are supposed to be for the tokens themselves instead of being converted into spans.
    This format makes sure that each tag is converted into a span of length ``1``. This lets us run metrics
    over individual tags without having to change our processing code. This is used for things like part of
    speech tagging.

    Due to the fact that there are no special prefixes for tokens that dictate the function a token plays
    in a span all the class values are left as ``None``.
    """


class SpanEncoding(Enum):
    """An enumeration of the kind of span encoding schemes we support processing."""

    TOKEN = TOKEN
    IOB = IOB
    BIO = BIO
    IOBES = IOBES
    BILOU = BILOU
    BMEOW = BMEOW
    BMEWO = BMEWO

    @classmethod
    def from_string(cls, value: str) -> "SpanEncoding":
        """Parse string into a specific span encoding format.

        Args:
            value: The string to dispatch to encoding on.

        Raises:
            ValueError: If the string cannot be recognized as pointing to a specific SpanEncoding format.

        Returns:
            The SpanEncoding member.
        """
        value = value.lower().strip()
        if value == "iob":
            return cls.IOB
        if value in ("iob2", "bio"):
            return cls.BIO
        if value == "iobes":
            return cls.IOBES
        if value in ("bilou", "bioul"):
            return cls.BILOU
        if value in ("bmewo", "bmeow"):
            return cls.BMEOW
        if value == "token":
            return cls.TOKEN
        raise ValueError(f"Unknown Encoding scheme, got: `{value}`")


class Span(NamedTuple):
    """Our representation of a span of text.

    Note:
        Our ``end`` attribute of a span is one greater than the index of the final token
        in the span. This is so that python list slicing works. For example,
        ``tokens[span.start : span.end]`` will yield the surface form of the span.

    Args:
        type: The type of the span in our downstream task, things like ``PER`` or ``LOC``.
        start: The index into the tokens list where the span starts.
        end: The index of the last token of the span plus 1.
        tokens: The indices that are part of the span.
    """

    type: str
    start: int
    end: int
    tokens: Tuple[int]


class ErrorType(Enum):
    pass


class Error(NamedTuple):
    """An error encountered when parsing tags into spans.

    Args:
        location: The index where the error occurred
        type: What kind of error is it. **TODO** These types need to be enumerated and hammer out the specifics
        current: The tag at the index of the error
        previous: The previous tag
        next: The next tag
    """

    location: int
    type: str
    current: str
    previous: str
    next: str

    def __str__(self):
        return f"{self.type} error at index {self.location}."


from iobes.convert import (
    iob_to_bio,
    iob_to_iobes,
    iob_to_bilou,
    iob_to_bmeow,
    iob_to_bmewo,
    bio_to_iob,
    bio_to_iobes,
    bio_to_bilou,
    bio_to_bmeow,
    bio_to_bmewo,
    iobes_to_iob,
    iobes_to_bio,
    iobes_to_bilou,
    iobes_to_bmeow,
    iobes_to_bmewo,
    bilou_to_iob,
    bilou_to_bio,
    bilou_to_iobes,
    bilou_to_bmeow,
    bilou_to_bmewo,
    bmeow_to_iob,
    bmeow_to_bio,
    bmeow_to_iobes,
    bmeow_to_bilou,
    bmewo_to_iob,
    bmewo_to_bio,
    bmewo_to_iobes,
    bmewo_to_bilou,
)
from iobes.parse import (
    parse_spans,
    parse_spans_token,
    parse_spans_iob,
    parse_spans_bio,
    parse_spans_iobes,
    parse_spans_bilou,
    parse_spans_bmeow,
    parse_spans_bmewo,
    parse_spans_with_errors,
    parse_spans_token_with_errors,
    parse_spans_iob_with_errors,
    parse_spans_bio_with_errors,
    parse_spans_iobes_with_errors,
    parse_spans_bilou_with_errors,
    parse_spans_bmeow_with_errors,
    parse_spans_bmewo_with_errors,
    validate_tags,
    validate_tags_iob,
    validate_tags_bio,
    validate_tags_iobes,
    validate_tags_bilou,
    validate_tags_bmeow,
    validate_tags_bmewo,
)
from iobes.transition import (
    Transition,
    transitions_legality,
    iob_transitions_legality,
    bio_transitions_legality,
    iobes_transitions_legality,
    bilou_transitions_legality,
    bmeow_transitions_legality,
    bmewo_transitions_legality,
    transitions_to_tuple_map,
    transitions_to_map,
)
from iobes.write import (
    write_tags,
    write_iob_tags,
    write_bio_tags,
    write_iobes_tags,
    write_bilou_tags,
    write_bmeow_tags,
    write_bmewo_tags,
)
from iobes.utils import (
    extract_type,
    extract_function,
)



from typing import List, Union, Any


def extract_type(tag: str, sep: str = "-") -> str:
    """Extract the span type from a tag.

    Tags are made of two parts. The second part is the type of the span which
    is specific to the downstream task. This function extracts that value from
    the tag.

    Args:
        tag: The tag to extract the type from.
        sep: The character (or string of characters) that separate the token
            function from the span type.

    Returns:
        The span type.
    """
    if sep not in tag:
        return tag
    return tag.split(sep, maxsplit=1)[1]


def extract_function(tag: str, sep: str = "-") -> str:
    """Extract the token function from a tag.

    Tags are made of two parts. The first part is the role that this tag plays
    in a span. It is generic across datasets (but differs across different span
    formatting options) and tells us things like this tag is the beginning or a
    span or this tag ends a span. This function extracts the token function or
    from the tag.

    Args:
        tag: The tag to extract the token function from.
        sep: The character (or string of characters) that separate the token
            function from the span type.

    Returns:
        The token function of this tag.
    """
    if sep not in tag:
        return tag
    return tag.split(sep, maxsplit=1)[0]


def safe_get(xs: List[Any], idx: int) -> Union[Any, None]:
    """Get the element at some index but return ``None`` when the index is out of bounds.

    Args:
        xs: The list to extract from.
        idx: The index to try to pull from.

    Returns:
        The value at ``idx`` or ``None`` if ``idx`` is out of bounds.
    """
    if idx < 0 or idx >= len(xs):
        return None
    return xs[idx]
